// derived from attachable_sstream_buf.hpp in boost.log, (C) 2007 Andrey Semashev
// modifications by Jason Wagner, 2008
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_LOGGING_DETAILS_LOG_STREAMBUF_HPP_
#define BOOST_LOGGING_DETAILS_LOG_STREAMBUF_HPP_

#include <memory>
#include <string>
#include <streambuf>

namespace boost { namespace logging { namespace details {

//! A streambuf that puts the formatted data to an external string
template<
    typename CharT,
    typename TraitsT = std::char_traits< CharT >,
    typename AllocatorT = std::allocator< CharT >
>
class log_streambuf :
    public std::basic_streambuf< CharT, TraitsT >
{
    //! Base type
    typedef std::basic_streambuf< CharT, TraitsT > base_type;

    //! Buffer size
    enum { buffer_size = 32 };

public:
    //! Character type
    typedef typename base_type::char_type char_type;
    //! Traits type
    typedef typename base_type::traits_type traits_type;
    //! String type
    typedef std::basic_string< char_type, traits_type, AllocatorT > string_type;
    //! Int type
    typedef typename base_type::int_type int_type;

private:
    //! A reference to the string that will be filled
    string_type m_Storage;
    //! A buffer used to temporarily store output
    char_type m_Buffer[buffer_size];

private:
    //! Copy constructor (closed)
  log_streambuf(log_streambuf const& that);
    //! Assignment (closed)
  log_streambuf& operator= (log_streambuf const& that);

public:
    //! Constructor
    explicit log_streambuf() : m_Storage()
    {
        base_type::setp(m_Buffer, m_Buffer + (sizeof(m_Buffer) / sizeof(*m_Buffer)));
    }

    string_type& str()
    {
      return m_Storage;
    }

    void clear_str()
    {
      m_Storage.clear();
    }

protected:
    //! Puts all buffered data to the string
    int sync()
    {
        register char_type* pBase = this->pbase();
        register char_type* pPtr = this->pptr();
        if (pBase != pPtr)
        {
            m_Storage.append(pBase, pPtr);
            this->pbump(static_cast< int >(pBase - pPtr));
        }
        return 0;
    }
    //! Puts an unbuffered character to the string
    int_type overflow(int_type c)
    {
      log_streambuf::sync();
        if (!traits_type::eq_int_type(c, traits_type::eof()))
        {
            m_Storage.push_back(traits_type::to_char_type(c));
            return c;
        }
        else
            return traits_type::not_eof(c);
    }
    //! Puts a character sequence to the string
    std::streamsize xsputn(const char_type* s, std::streamsize n)
    {
      log_streambuf::sync();
        typedef typename string_type::size_type string_size_type;
        register const string_size_type max_storage_left =
            m_Storage.max_size() - m_Storage.size();
        if (static_cast< string_size_type >(n) < max_storage_left)
        {
            m_Storage.append(s, static_cast< string_size_type >(n));
            return n;
        }
        else
        {
            m_Storage.append(s, max_storage_left);
            return static_cast< std::streamsize >(max_storage_left);
        }
    }
};

}}} // namespace boost::logging::details


#endif // BOOST_LOGGING_DETAILS_LOG_STREAMBUF_HPP_
