//          Copyright Jason Wagner 2008
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_LOG_DETAILS_MESSAGE_HPP_
#define BOOST_LOG_DETAILS_MESSAGE_HPP_
#include <boost/logging/details/message_helper.hpp>
#include <boost/logging/manipulator.hpp>

namespace boost { namespace logging { namespace details {
/* The message class implements the static elimination of output
 * based upon the current tag types.
 * For a given series of insertions to a logger:
 * logger << a << b << c << end_record();
 * each insertion returns a new message<> that properly evaluates whether
 * the next records should be logged.  There are three basic types of things
 * that get inserted-- things to output (TYPE and ostream_functions), things
 * to do to the logger (manipulator), and tags.
 *
 * Message has 4 template parameters-
 *      IsActive -- whether insertions will have an effect
 *      TagMap -- What are the current tags in effect?
 *      LoggerType -- the type of logger we're inserting into
 *      IsRecordOpen  -- have we even started the record yet.
 *
 * IsActive
 *   message is partially specialized on this.  If it is false, then << on output
 *   and manipulators are empty functions and optimized away.  If it is true, they
 *   are processed.  Only << tag can change the state of this flag.
 * TagMap
 *   This is a map of tags to their values (mpl::map).  It is used to determine switches
 *   in IsActive when the LoggerType::ConditionMap is applied to them.  <<(Tag<>) inserts
 *   into this map and causes re-evaluation of IsActive.
 * LoggerType
 *   Used for storing the logger to feed this information to, as well as for string types,
 *   ostream types, static conditions, etc.
 * IsRecordOpen
 *   Is false if start_record() needs to be called, true if there is an open record.  In the
 *   case that a given line has no effective output, this prevents a record and the
 *   subsequent expensive get_sinks() call from ever being made.
 *
 */


/* Rules for is_open_record:
 *  << Tag -- can never open a record
 *  << Manipulator -- can never open a record
 *  << ostream_function -- opens a record if is_active
 *  << TYPE -- opens a record if is_active
 */

/**
 * Specialization for isActive==true, where the functions actually do stuff.
 */
// consider partial specialization so that this can return *this as record_open_type in
// situations where that makes sense?
template<class TagMap, class LoggerType, class IsRecordOpen>
struct message<boost::mpl::true_,TagMap,LoggerType,IsRecordOpen>
{
  typedef TagMap tag_map;
  typedef LoggerType logger_type;
  typedef boost::mpl::true_ is_active;
  typedef typename LoggerType::char_type char_type;
  typedef typename LoggerType::traits_type traits_type;
  typedef message<boost::mpl::true_,TagMap,LoggerType,IsRecordOpen> self_type;
  typedef message<boost::mpl::true_,TagMap,LoggerType,is_active> record_open_type;

  typedef std::basic_ostream<char_type,traits_type> ostream_type;

  message(LoggerType& s) :logger_(s) {}

  template<class TYPE>
  record_open_type operator<<(const TYPE& t)
  {
    if(!IsRecordOpen())
      logger_.template start_record<tag_map>();

    logger_.write(t);
    return record_open_type(logger_);
  }

  template<class Category, int Value>
  typename message_return<Category,Value,TagMap,LoggerType,IsRecordOpen>::type
  operator<<(tag<Category,Value> t)
  {
    typedef typename message_return<Category,Value,TagMap,LoggerType,IsRecordOpen>::type rettype;
    if(IsRecordOpen::value)
      logger_.record_varset().set(Category().type_string(),Value);
    return  rettype(logger_);
  }

  template<class Op>
  self_type& operator<<(const manipulator<Op>& t)
  {
    static_cast<const Op&>(t)(logger_,true);
    return *this;
  }

  record_open_type operator<<(ostream_type& (*f)(ostream_type&))
  {
    if(!IsRecordOpen())
      logger_.template start_record<tag_map>();
    logger_.write(f);
    return record_open_type(logger_);
  }


protected:
  LoggerType& logger_;
};

/**
 * Specialization for isActive=false, where functions do nothing except eat input and
 * move on to the next step.
 */
template<class TagMap, class LoggerType, class IsRecordOpen>
struct message<boost::mpl::false_,TagMap,LoggerType,IsRecordOpen>
{
  typedef TagMap tag_map;
  typedef LoggerType logger_type;
  typedef boost::mpl::false_ is_active;
  typedef typename LoggerType::char_type char_type;
  typedef typename LoggerType::traits_type traits_type;
  typedef message<boost::mpl::false_,TagMap,LoggerType,IsRecordOpen> self_type;
  typedef std::basic_ostream<char_type,traits_type> ostream_type;

  message(LoggerType& s) : logger_(s) {}
  template<class TYPE>
  self_type& operator<<(TYPE)
  {
    return *this;
  }

  template<class Category, int Value>
  typename message_return<Category,Value,TagMap,LoggerType,IsRecordOpen>::type
  operator<<(tag<Category,Value> t)
  {
    typedef typename message_return<Category,Value,TagMap,LoggerType,IsRecordOpen>::type rettype;
    if(IsRecordOpen::value)
      logger_.record_varset().set(Category().type_string(),Value);
    return rettype(logger_);
  }

  template<class Op>
  self_type& operator<<(const manipulator<Op>& t)
  {
    if(IsRecordOpen::value)
      static_cast<const Op&>(t)(logger_,false);
    return *this;
  }

  self_type& operator<<(ostream_type& (*f)(ostream_type&))
  {
    return *this;
  }

  protected:
  LoggerType& logger_;
};

}}}; //namespace boost::logging::details

#endif /*BOOST_LOGGING_DETAILS_MESSAGE_HPP_*/
