//          Copyright Jason Wagner 2008
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_LOGGING_FILTERS_HPP_
#define BOOST_LOGGING_FILTERS_HPP_
#include <set>
#include <boost/shared_ptr.hpp>
#include <boost/utility.hpp>
#include <boost/logging/details/variable_comparisons.hpp>
#include <boost/logging/tag.hpp>
#include <boost/logging/variable.hpp>

namespace boost { namespace logging {

struct sink;

/**
 * A holder to make reference to a variable in a varset.  Used
 * for the filter generators.
 */
struct var {
  explicit var(const std::string& s) : type(s) {}
  const std::string& type_string() const { return type;}
private:
  std::string type;
};

/**
 * Abstract base class for all filters.  Users shouldn't have to deal
 * with this too much, but you can use the filter_base::ptr to point to
 * intermediate results:
 * filter_base::ptr f= severity() < severity::debug() && build_type() == build_type::debug()
 */
struct filter_base {
  typedef boost::shared_ptr<filter_base> ptr;
  virtual bool eval(const variable_set& v) const=0;
};

/**
 * Template for the operation comparing a var to var:
 * var("my_variable") == var("my_other_variable")
 */
template<class Operation>
struct binary_variable_filter : public filter_base {
  typedef default_variable_type value_type;
  typedef Operation operation;
  binary_variable_filter(const std::string& lhs, const std::string& rhs)
    : lhs_(lhs),rhs_(rhs)
  {
  }

  bool eval(const variable_set& v) const
  {
    return boost::apply_visitor(operation(),v.get(lhs_),v.get(rhs_));
  }
private:
  std::string lhs_;
  std::string rhs_;
};

/**
 * Template for the comparison of a var to a constant variable:
 * var("my_variable") == "foo"
 * severity() == severity::debug()
 */
template<class Operation>
struct binary_constant_filter : public filter_base {
  typedef default_variable_type value_type;
  typedef Operation operation;
  binary_constant_filter(const std::string& lhs, const value_type& rhs)
    : lhs_(lhs),rhs_(rhs)
  {
  }

  bool eval(const variable_set& v) const
  {
//    std::cerr << "Comparing " << lhs_ << " lhs=" << v.get(lhs_) << " to rhs=" << rhs_ << "\n";
    return boost::apply_visitor(operation(),v.get(lhs_),rhs_);
  }
private:
  std::string lhs_;
  value_type rhs_;
};

typedef binary_variable_filter<details::equals>        variable_equals_filter;
typedef binary_variable_filter<details::not_equals>    variable_not_equals_filter;
typedef binary_variable_filter<details::less_than>     variable_less_than_filter;
typedef binary_variable_filter<details::less_equal>    variable_less_equal_filter;
typedef binary_variable_filter<details::greater_than>  variable_greater_than_filter;
typedef binary_variable_filter<details::greater_equal> variable_greater_equal_filter;

typedef binary_constant_filter<details::equals>        const_equals_filter;
typedef binary_constant_filter<details::not_equals>    const_not_equals_filter;
typedef binary_constant_filter<details::less_than>     const_less_than_filter;
typedef binary_constant_filter<details::less_equal>    const_less_equal_filter;
typedef binary_constant_filter<details::greater_than>  const_greater_than_filter;
typedef binary_constant_filter<details::greater_equal> const_greater_equal_filter;

struct and_filter : public filter_base {
  and_filter(filter_base::ptr lhs, filter_base::ptr rhs)
    : lhs_(lhs),rhs_(rhs)
  {
  }

  bool eval(const variable_set& v) const
  {
    return lhs_->eval(v) && rhs_->eval(v);
  }
private:
  filter_base::ptr lhs_;
  filter_base::ptr rhs_;
};

struct or_filter : public filter_base {
  or_filter(filter_base::ptr lhs, filter_base::ptr rhs)
    : lhs_(lhs),rhs_(rhs)
  {
  }

  bool eval(const variable_set& v) const
  {
    return lhs_->eval(v) || rhs_->eval(v);
  }
private:
  filter_base::ptr lhs_;
  filter_base::ptr rhs_;
};

struct not_filter : public filter_base {
  not_filter(filter_base::ptr lhs)
    : lhs_(lhs)
  {
  }

  bool eval(const variable_set& v) const
  {
    return !lhs_->eval(v) ;
  }
private:
  filter_base::ptr lhs_;
};

// using a tag
//filter f1= bl::severity() > bl::severity::debug();
template<class Tag,int Value>
inline filter_base::ptr operator<(const Tag& t, const tag<Tag,Value>& v)
{
  return shared_ptr<filter_base>( new const_less_than_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator<(const tag<Tag,Value>& v,const Tag& t)
{
  return filter_base::ptr(new  const_greater_equal_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator<=(const Tag& t, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new const_less_equal_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator<=(const tag<Tag,Value>& v,const Tag& t)
{
  return filter_base::ptr(new  const_greater_than_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator>(const Tag& t, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new  const_greater_than_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator>(const tag<Tag,Value>& v,const Tag& t)
{
  return filter_base::ptr(new  const_less_equal_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator>=(const Tag& t, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new  const_greater_equal_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator>=(const tag<Tag,Value>& v,const Tag& t)
{
  return filter_base::ptr(new  const_less_than_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator==(const Tag& t, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new  const_equals_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator==(const tag<Tag,Value>& v,const Tag& t)
{
  return filter_base::ptr(new  const_equals_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator!=(const Tag& t, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new  const_not_equals_filter(t.type_string(),Value));
}
template<class Tag,int Value>
inline filter_base::ptr operator!=(const tag<Tag,Value>& v,const Tag& t)
{
  return filter_base::ptr(new  const_not_equals_filter(t.type_string(),Value));
}
// using a variable type
//filter f2= bl::file_name() == "foo";
template<class ValueType>
inline filter_base::ptr operator<(const variable& type, const ValueType& v)
{
  return filter_base::ptr(new  const_less_than_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator<(const ValueType& v,const variable& type)
{
  return filter_base::ptr(new  const_greater_equal_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator<=(const variable& type, const ValueType& v)
{
  return filter_base::ptr(new  const_less_equal_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator<=(const ValueType& v,const variable& type)
{
  return filter_base::ptr(new const_greater_than_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator>(const variable& type, const ValueType& v)
{
  return filter_base::ptr(new const_greater_than_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator>(const ValueType& v,const variable& type)
{
  return filter_base::ptr(new const_less_equal_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator>=(const variable& type, const ValueType& v)
{
  return filter_base::ptr(new const_greater_equal_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator>=(const ValueType& v,const variable& type)
{
  return filter_base::ptr(new const_less_than_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator==(const variable& type, const ValueType& v)
{
  return filter_base::ptr(new const_equals_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator==(const ValueType& v,const variable& type)
{
  return filter_base::ptr(new const_equals_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator!=(const variable& type, const ValueType& v)
{
  return filter_base::ptr(new const_not_equals_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator!=(const ValueType& v,const variable& type)
{
  return filter_base::ptr(new const_not_equals_filter(type.type_string(),v));
}
// using a variable constant
//filter f3= bl::var("file_name") == "foo";
template<class ValueType>
inline filter_base::ptr operator<(const var& type, const ValueType& v)
{
  return filter_base::ptr(new const_less_than_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator<(const ValueType& v,const var& type)
{
  return filter_base::ptr(new const_greater_equal_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator<=(const var& type, const ValueType& v)
{
  return filter_base::ptr(new const_less_equal_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator<=(const ValueType& v,const var& type)
{
  return filter_base::ptr(new const_greater_than_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator>(const var& type, const ValueType& v)
{
  return filter_base::ptr(new const_greater_than_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator>(const ValueType& v,const var& type)
{
  return filter_base::ptr(new const_less_equal_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator>=(const var& type, const ValueType& v)
{
  return filter_base::ptr(new const_greater_equal_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator>=(const ValueType& v,const var& type)
{
  return filter_base::ptr(new const_less_than_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator==(const var& type, const ValueType& v)
{
  return filter_base::ptr(new const_equals_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator==(const ValueType& v,const var& type)
{
  return filter_base::ptr(new const_equals_filter(type.type_string(),v));

}
template<class ValueType>
inline filter_base::ptr operator!=(const var& type, const ValueType& v)
{
  return filter_base::ptr(new const_not_equals_filter(type.type_string(),v));
}
template<class ValueType>
inline filter_base::ptr operator!=(const ValueType& v,const var& type)
{
  return filter_base::ptr(new const_not_equals_filter(type.type_string(),v));
}
// a variable versus a tag (compares the int constant)
//filter f4= bl::var("my_severity") < bl::severity::debug();
template<class Tag, int Value>
inline filter_base::ptr operator<(const var& type, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new const_less_than_filter(type.type_string(),Value));
}
template<class Tag, int Value>
inline filter_base::ptr operator<(const tag<Tag,Value>& v,const var& type)
{
  return filter_base::ptr(new const_greater_equal_filter(type.type_string(),Value));
}
template<class Tag, int Value>
inline filter_base::ptr operator<=(const var& type, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new const_less_equal_filter(type.type_string(),Value));
}
template<class Tag, int Value>
inline filter_base::ptr operator<=(const tag<Tag,Value>& v,const var& type)
{
  return filter_base::ptr(new const_greater_than_filter(type.type_string(),Value));
}
template<class Tag, int Value>
inline filter_base::ptr operator>(const var& type, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new const_greater_than_filter(type.type_string(),Value));
}
template<class Tag, int Value>
inline filter_base::ptr operator>(const tag<Tag,Value>& v,const var& type)
{
  return filter_base::ptr(new const_less_equal_filter(type.type_string(),Value));
}
template<class Tag, int Value>
inline filter_base::ptr operator>=(const var& type, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new const_greater_equal_filter(type.type_string(),Value));
}
template<class Tag, int Value>
inline filter_base::ptr operator>=(const tag<Tag,Value>& v,const var& type)
{
  return filter_base::ptr(new const_less_than_filter(type.type_string(),Value));
}
template<class Tag, int Value>
inline filter_base::ptr operator==(const var& type, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new const_equals_filter(type.type_string(),v));
}
template<class Tag, int Value>
inline filter_base::ptr operator==(const tag<Tag,Value>& v,const variable& type)
{
  return filter_base::ptr(new const_equals_filter(type.type_string(),v));
}

template<class Tag, int Value>
inline filter_base::ptr operator!=(const variable& type, const tag<Tag,Value>& v)
{
  return filter_base::ptr(new  variable_not_equals_filter(type.type_string(),v));
}
template<class Tag, int Value>
inline filter_base::ptr operator!=(const tag<Tag,Value>& v,const variable& type)
{
  return filter_base::ptr(new  variable_not_equals_filter(type.type_string(),v));
}


// var vs var
// filter f5= var("my_var") >= var("my_global_var")
// TODO: var() versus var() filter generators

// two filters, logical ops
inline filter_base::ptr operator&&(filter_base::ptr lhs, filter_base::ptr rhs)
{
  return filter_base::ptr(new and_filter(lhs,rhs));
}
inline filter_base::ptr operator||(filter_base::ptr lhs, filter_base::ptr rhs)
{
  return filter_base::ptr(new or_filter(lhs,rhs));
}
inline filter_base::ptr operator!(filter_base::ptr lhs)
{
  return filter_base::ptr(new not_filter(lhs));
}
//=============================================================================
// Filter actions
//=============================================================================

// Filters are conditions, these are the actual actions for a filter

/**
 * The basic interface for filter_actions
 */
struct filter_action : boost::noncopyable
{
  typedef std::vector<sink*> sink_list;
  typedef boost::shared_ptr<filter_action> ptr;

  virtual ~filter_action() {}
  /**
   * Returns all the sinks that are interested based upon this variable set.
   * @param v
   * @param s
   */
  virtual void get_sinks(const variable_set& v, sink_list& s) =0;

};

/**
 * A splitting filter action for chaining multiple actions based upon a condition.
 * The condition is a filter.  The split has three basic forks:
 *   on_success -- if the filter returns true
 *   on_failure -- if the filter returns false
 *   always -- regardless of what the filter returns
 * This allows routing of log records, such as sending everything of severity::warn or above
 * to a scrolling console while anything less to a rolling log file and everything to a main log:
 * @code
 * filter_split_action f(severity() >= severity::warn());
 * f.on_success(my_console_log);
 * f.on_failure(my_rolling_debug_log_file);
 * f.always(my_main_log_file);
 * @endcode
 *
 * There can be multiple filter_actions for each branch, the order called is undefined.
 */
struct filter_split_action : public filter_action {
  typedef boost::shared_ptr<filter_split_action> ptr;

  filter_split_action() {}
  filter_split_action(filter_base::ptr p) : condition(p) {}

  /**
   *
   * @param v
   * @param s
   */
  virtual void get_sinks(const variable_set& v, sink_list& s)
  {
    bool success=condition->eval(v);
    for(filter_set::iterator fI=always_.begin(); fI != always_.end(); ++fI)
      (*fI)->get_sinks(v,s);

    if(success)
    {
      for(filter_set::iterator fI=on_success_.begin(); fI != on_success_.end(); ++fI)
        (*fI)->get_sinks(v,s);
    } else {
      for(filter_set::iterator fI=on_failure_.begin(); fI != on_failure_.end(); ++fI)
        (*fI)->get_sinks(v,s);
    }
  }
  void set_filter(const filter_base::ptr ptr)
  {
    condition=ptr;
  }
  void always(filter_action::ptr f)
  {
    always_.insert(f);
  }
  void on_success(filter_action::ptr f)
  {
    on_success_.insert(f);
  }
  void on_failure(filter_action::ptr f)
  {
    on_failure_.insert(f);
  }

  void remove(filter_action::ptr f)
  {
    always_.erase(f);
    on_success_.erase(f);
    on_failure_.erase(f);
  }
  void remove_always(filter_action::ptr f)
  {
    always_.erase(f);
  }
  void remove_on_success(filter_action::ptr f)
  {
    on_success_.erase(f);
  }
  void remove_on_failure(filter_action::ptr f)
  {
    on_failure_.erase(f);
  }

private:
  typedef std::set<filter_action::ptr> filter_set;
  filter_set on_success_;
  filter_set on_failure_;
  filter_set always_;

  filter_base::ptr condition;

};

}} // namespace boost::logging

#endif /* BOOST_LOGGING_FILTERS_HPP_ */
