//          Copyright Jason Wagner 2008
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_LOGGING_LOGGER_HPP_
#define BOOST_LOGGING_LOGGER_HPP_

#include <vector>
#include <map>
#include <boost/noncopyable.hpp>
#include <boost/function.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/mpl/map.hpp>
#include <boost/mpl/logical.hpp>
#include <boost/mpl/has_key.hpp>
#include <boost/mpl/for_each.hpp>
#include <boost/logging/tag.hpp>
#include <boost/logging/details/message_helper.hpp>
#include <boost/logging/details/message.hpp>
#include <boost/logging/manipulator.hpp>
#include <boost/logging/conditions.hpp>
#include <boost/logging/variable.hpp>
#include <boost/logging/filters.hpp>
#include <boost/logging/single_threaded_variable_set.hpp>
#include <boost/logging/sinks/sink.hpp>
#include <boost/logging/details/log_streambuf.hpp>

namespace boost { namespace logging {

typedef boost::mpl::vector<> default_conditions;
typedef boost::mpl::vector<> default_tags;

typedef boost::mpl::vector<
    condition::greater_equal<severity::warning>,
    condition::in<build_type::release>
    > optimized_conditions;

typedef boost::mpl::vector<
    condition::in<build_type::debug>
    > debug_conditions;

typedef boost::mpl::vector<> empty_conditions;

typedef boost::mpl::vector<build_type::release,severity::info> optimized_tags;
typedef boost::mpl::vector<build_type::debug,severity::info> debug_tags;

#if defined(_DEBUG) || !defined(NDEBUG)
  #define BOOST_LOG_DEFAULT_CONDITIONS boost::logging::debug_conditions
  #define BOOST_LOG_DEFAULT_TAGS boost::logging::debug_tags
#else
  #define BOOST_LOG_DEFAULT_CONDITIONS boost::logging::optimized_conditions
  #define BOOST_LOG_DEFAULT_TAGS boost::logging::optimized_tags
#endif

#if defined(BOOST_LOG_USE_WCHAR)
  #define BOOST_LOG_OSTRINGSTREAM ::std::wostringstream
#else
  #define BOOST_LOG_OSTRINGSTREAM ::std::ostringstream
#endif

struct null_deleter
{
    void operator()(void const *) const  {}
};

/**
 * Logger is the entry point into the system.
 * Template Parameters:
 *    TagConditions
 *      An mpl::vector<> of boost::logging::condition::* types to determine at compile time whether to log
 *      based upon tags inserted by the user.  All conditions are "anded" together.  If the tag is not defined,
 *      then it matches all conditions.
 *      Defaults to condition::greater_equal<severity::warning>, condition::in<build_type::release> for optimized builds
 *      Defaults to condition::in<build_type::debug> for debug builds.
 *    DefaultTagList
 *      An mpl::vector<> of tags that are defined for the logger before the first insertion occurs.
 *      Defaults to <build_type::release, severity::info> for optimized builds
 *      Defaults to <build_type::debug, severity::info> for debug builds
 *    CharT, TraitsT
 *      Support wchar_t and i19n.  Not really used at the moment.
 *
 *  Notes:
 *    Generally use the filter_action::ptr& constructor.  A logger with no action doesn't
 *    really do anything.
 *
 *    The functions write() and start_record() should be considered private.  They are public
 *    because friend with templates of this type is prohibitively hard.
 *
 *
 *
 */
template<class TagConditions=BOOST_LOG_DEFAULT_CONDITIONS,
         class DefaultTagList=BOOST_LOG_DEFAULT_TAGS,
         class CharT = char,
         class TraitsT= std::char_traits< CharT >
         >
struct logger : boost::noncopyable
{
  typedef TagConditions ConditionMap;
  typedef typename boost::mpl::fold<DefaultTagList, boost::mpl::map<> , details::tag_vector_to_map>::type DefaultTagMap;
  typedef CharT char_type;
  typedef TraitsT traits_type;

  typedef typename boost::logging::details::message_eval<TagConditions,DefaultTagMap>::type is_active_default;
  typedef boost::logging::details::log_streambuf<char_type,traits_type> stream_buf_type;
  typedef ::std::basic_ostream<char_type,traits_type> ostream_type;
  typedef ::std::basic_string<char_type,traits_type> string_type;
  typedef logger<TagConditions,DefaultTagList,CharT,TraitsT> self_type;

  logger() : out_(&stream_buf_) {}
  logger(filter_action::ptr filter) : filter_(filter), out_(&stream_buf_)
  {
    variable_set::ptr p(&logger_var_set_,null_deleter());
    record_var_set_.set_parent(p);
    dirty_vars_=true;
  }

  //---------------------------------------------------------------------------
  /**@name Writer functions
   * These functions actually write text to the streams after all the static
   * conditions have been satisfied.
  */
  //@{
  /**
   * This handles the general case, anything that ostream handles should be handled by this.
   * @param t An arbitrary type that ostream supports
   */
  template<class TYPE> void write(TYPE t)
  {
    if(dirty_vars_)
        reload_sinks();
    if(!active_builders_.empty())
      out_ << t;
  }

  /**
   * This handles lazy evaluation of potentially long functions via boost::function.  Only
   * handles nullary functions right now, but could be extended easily if needed to accept the varset.
   * @param t - a nullary function to be evaluated
   */
  template<class RV>
    void write(const boost::function<RV ()>& t)
  {
    if(dirty_vars_)
        reload_sinks();
    if(!active_builders_.empty())
      out_ << t();
  }
  //@}

  //---------------------------------------------------------------------------
  /**@name Operators <<
   * These overrides exist to eat the first argument inserted into the logger.
   * They kickstart the message<> chain.  The overload order occurs based upon
   * the specificity of the param-- tag, manipulator, and TYPE come first, since
   * they are templates.  tag and manipulator are chosen over TYPE because they
   * are more specific.  ostream operator occurs last because the function is
   * ambiguous for char and wchar in the templated versions.
  */
  //@{
  typedef typename boost::logging::details::message<
     typename boost::logging::details::message_eval<TagConditions,DefaultTagMap>::type,
                                  DefaultTagMap,
                                  self_type,
                                  is_active_default
     > untagged_return_type;

  /**
   * Eats all types not caught by other operators
   * @param t An arbitrary type supported by ostream
   * @return a message object to continue the insertion chain
   */
  template<class TYPE>
  untagged_return_type operator<<(TYPE t)
  {
    if(is_active_default::value)
    {
      start_record<DefaultTagMap>();
      write(t);
    }
    return untagged_return_type(*this);
  }

  /**
   * Catches all iostream manipulators due to ambiguity between
   * basic_ostream<char> and basic_ostream<wchar> when the <<(TYPE t) overload tries
   * to eat the type.
   * @param f An ostream manipulator
   * @return A message object to continue the insertion chain
   */
  untagged_return_type operator<<(ostream_type& (*f)(ostream_type&))
  {
    if(is_active_default::value)
    {
      start_record<DefaultTagMap>();
      write(f);
    }
    return untagged_return_type(*this);
  }

  /**
   * Calls a logger manipulator.
   * @param t the manipulator
   * @return A message object to continue the insertion chain.
   */
  template<class Op>
  untagged_return_type operator<<(const manipulator<Op>& t)
  {
    static_cast<const Op&>(t)(*this,is_active_default::value);
    return untagged_return_type(*this);
  }

  /**
   * Inserts a tag, affecting the static conditions for logging.  Does no
   * work on it's own, but builds a message<> type that may allow or disallow
   * future insertions from logging.
   * @param t
   * @return
   */
  template<class Category, int Value>
    typename boost::logging::details::message_return<
      Category,Value,DefaultTagMap,self_type,boost::mpl::false_
    >::type
    operator<<(boost::logging::tag<Category,Value> t)
  {
    typedef typename boost::logging::details::message_return<
       Category,Value,DefaultTagMap,self_type,boost::mpl::false_
    >::type  rettype;
    typedef typename rettype::is_active is_active;
    return rettype(*this);
  }
  //@}

  //---------------------------------------------------------------------------
  /**@name Record Management
  */
  //@{
  /**
   * Begins a new record.  Takes a tag map as a parameter, which it populates
   * the logger's local variables with.  This allows for the static tagging to
   * be truly static and only populated for filters at the last possible moment.
   * SHOULD ONLY BE CALLED INTERNALLY
   */
  template<class CurrentTagMap>
  void start_record()
  {
    dirty_vars_=true;
    Setter setter(record_var_set_);
    boost::mpl::for_each<CurrentTagMap>(setter);
  }

  /**
   * Ends a record and writes it to all output streams.  It also resets the record
   * for the next logger output.
   */
  void end_record()
  {
    flush();
    for(typename sink_map::iterator tI=sink_map_.begin(); tI != sink_map_.end(); ++tI)
    {
      tI->first->write(tI->second);
    }
    sink_map_.clear();
    active_builders_.clear();
    record_var_set_.clear();
  }
  //@}

  /**
   * Gets the logger's local veriables and marks them dirty.
   * @return
   */
  variable_set& local()
  {
    dirty_vars_=true;
    return logger_var_set_;
  }

  /**
   * Gets the varset for the current record.  This gets cleared on
   * every end_record().  It's used for tags and set_variable type manipulators
   * inserted into the logger/message.
   * @return
   */
  variable_set& record_varset()
  {
    dirty_vars_=true;
    return record_var_set_;
  }

  /**
   * Sets the action performed by this logger.
   * @param filter
   */
  void set_action(const filter_action::ptr& filter)
  {
    filter_ = filter;
    dirty_vars_=true;
  }

private:
  void flush()
  {
    for(typename active_list::iterator itI=active_builders_.begin(); itI != active_builders_.end();++itI)
      (*itI)->second.append(stream_buf_.str());
    stream_buf_.clear_str();
  }

  // functor to translate tag types to variables
  struct Setter {
    single_threaded_variable_set& var;
    Setter(single_threaded_variable_set& l) : var(l) {}
    template<class TYPE>
    void operator()(const TYPE& t)
    {
      typedef typename TYPE::first cat;
      typedef typename TYPE::second val;
      var.set(cat().type_string(),val::value);
    }
  };
  // for debugging
  struct Printer {
    single_threaded_variable_set& var;
    Printer(single_threaded_variable_set& l) : var(l) {}
    template<class TYPE>
    void operator()(const TYPE& t)
    {
      typedef typename TYPE::first cat;
      typedef typename TYPE::second val;
//      std::cout << " Value: " << cat().type_string() << " == " << val::value << "\n";
    }
  };

  void reload_sinks()
  {
      dirty_vars_=false;
      // dump the currently built string to all the appropriate strings
      flush();

      // figure out new sinks that are interested in this message
      sinks_.clear();
      active_builders_.clear();
      filter_->get_sinks(record_var_set_,sinks_);
      if(sinks_.empty())
        return;

      for(typename filter_action::sink_list::iterator itI=sinks_.begin();itI!=sinks_.end();++itI)
      {
        typename sink_map::iterator tok=sink_map_.find(*itI);
        if(tok == sink_map_.end())
          tok =sink_map_.insert(typename sink_map::value_type(*itI,string_type())).first;
        active_builders_.push_back(tok);
      }
  }

//-------------------------------------------------------------------------------------------------
// Affected by the channel.
  filter_action::ptr filter_;
  single_threaded_variable_set logger_var_set_;

//-------------------------------------------------------------------------------------------------
// Local to this logger for building records.
  single_threaded_variable_set record_var_set_;

  // used to build the record before passing to the sinks
  stream_buf_type stream_buf_;
  ostream_type out_;

  // Build the messages that each sink will receive
  typedef std::map<sink*,string_type> sink_map;
  typedef std::vector<typename sink_map::iterator> active_list;
  active_list active_builders_; // who is currently receiving?
  sink_map sink_map_;  // maps everyone we've seen to the buffer they will receive

  // Used to fetch active sinks.  Member so we don't have to pay for it's construction.
  filter_action::sink_list sinks_;

  // Variables have changed, so recalculate the active sinks
  bool dirty_vars_;

};
}} // namespace boost::loglogging


#endif /*BOOST_LOGGING_LOGGER_HPP_*/
