#include <iostream>
#include <time.h>
#include <iomanip>
#include <boost/logging/logger.hpp>
#include <boost/logging/conditions.hpp>
#include <boost/logging/sinks/ostream_sink.hpp>

namespace bl=boost::logging;

const int COUNT=1000000;  // for timing test
//const int COUNT=10000;  // for profiling
//const int COUNT=1;      // for eyeball debugging

// Most of these functions are templated on the logger type to make changing the logger template params
// much easier.

// use the logger with the first line skipped, further lines printed
template<class L>
void test_logger(L& s)
{
  clock_t start, end;
  start=clock();
  for(int i=0; i<COUNT;++i)
  {
    s << "Default<Test> " << " default<fail>\n"
      << bl::severity::warning() << "warn(1234) " << "warn(asfd) warn(f=" <<21.1234678 << ")\n"
      << bl::severity::error() << "error(i=" << 45 << ") error (text here)\n";
    s.end_record();
  }
  end=clock();
  std::cerr << "Logging took " << double(end-start)/double(CLOCKS_PER_SEC) << "\n";
}

void test_cout()
{
  clock_t start, end;
  start=clock();
  for(int i=0; i<COUNT;++i)
  {
    std::cout << "Default<Test> " << " default<fail>"
      << "warn(1234) " << "warn(asfd) warn(f=" <<21.1234678 << ")\n"
      << "error(i=" << 45 << ") error (text here)\n";
  }
  end=clock();
  std::cerr << "std::cout took " << double(end-start)/double(CLOCKS_PER_SEC) << "\n";
}

void test_log_streambuf()
{
  clock_t start, end;
  start=clock();
  bl::details::log_streambuf<char> buffer;
  std::ostream out(&buffer);
  for(int i=0; i<COUNT;++i)
  {
    out << "Default<Test> " << " default<fail>"
      << "warn(1234) " << "warn(asfd) warn(f=" <<21.1234678 << ")\n"
      << "error(i=" << 45 << ") error (text here)\n";
//    std::cout << out.str();
    buffer.clear_str();
  }
  end=clock();
  std::cerr << "log_streambuf took " << double(end-start)/double(CLOCKS_PER_SEC) << "\n";
}

template<class L>
void test_all_accepted(L& s)
{
  clock_t start, end;
  start=clock();
  for(int i=0; i<COUNT;++i)
  {
    s << bl::severity::error() << "Default<Test> " << " default<fail>"
      << "warn(1234) " << "warn(asfd) warn(f=" <<21.1234678 << ")\n"
      << "error(i=" << 45 << ") error (text here)\n";
    s.end_record();
  }
  end=clock();
  std::cerr << "All_accepted took " << double(end-start)/double(CLOCKS_PER_SEC) << "\n";
}

template<class L>
void test_all_rejected(L& s)
{
  clock_t start, end;
  start=clock();
  for(int i=0; i<COUNT;++i)
  {
    s << bl::severity::info() << "Default<Test> " << " default<fail>"
    << "warn(1234) " << "warn(asfd) warn(f=" <<21.1234678 << ")\n"
    << "error(i=" << 45 << ") error (text here)\n";
    s.end_record();
  }
  end=clock();
  std::cerr << "All_rejected took " << double(end-start)/double(CLOCKS_PER_SEC) << "\n";
}

template<class L>
void test_all_rejected_compile_time(L& s)
{
  clock_t start, end;
  start=clock();
  for(int i=0; i<COUNT;++i)
  {
    s << bl::severity::debug() << "Default<Test> " << " default<fail>"
    << "warn(1234) " << "warn(asfd) warn(f=" <<21.1234678 << ")\n"
    << "error(i=" << 45 << ") error (text here)\n";
    s.end_record();
  }
  end=clock();
  std::cerr << "All_rejected at compile time took " << double(end-start)/double(CLOCKS_PER_SEC) << "\n";
}

// a worker function to just eat time.
long sum(long n)
{
  if(n==0)
    return 0;
  return n+sum(n-1);
}

// not really empty, but this makes the loop actually run rather than just go poof in the optimizer
template<class L>
void test_empty_loop(L& s)
{
  clock_t start, end;
  start=clock();
  long a=1;
  for(int i=0; i<COUNT;++i)
  {
    a+=sum(i&1);
  }
  end=clock();
  std::cerr << "Empty loop took " << double(end-start)/double(CLOCKS_PER_SEC) << " and produced " << a << "\n";
}

// MSVC eliminates sum(), gcc does not.  was just checking
template<class L>
void test_dead_code_elimination(L& s)
{
  clock_t start, end;
  start=clock();
  for(int i=0; i<COUNT;++i)
  {
    s << bl::severity::debug() << "sum function=" << sum(50000) << ")\n"
//      << bl::severity::warning() << "warn(1234) " << "warn(asfd) warn(f=" << " , " << 21.1234678 << ")\n"
//      << bl::severity::error() << "error(i=" << 45 << ") error (text here)\n";
    ;
    s.end_record();
  }
  end=clock();
  std::cerr << "DCE  took " << double(end-start)/double(CLOCKS_PER_SEC) << "\n";
}
int main()
{
  typedef boost::mpl::vector<
    bl::condition::greater<bl::severity::debug>
      > conditions;

  bl::filter_base::ptr f=bl::severity() > bl::severity::info();
  bl::filter_split_action::ptr fcl(new bl::filter_split_action(f));
  bl::sink::ptr sink(new bl::sinks::ostream_sink(std::cout));
  fcl->on_success(sink);

  // optimized tags = build_type::release,severity::info
  bl::logger<conditions,bl::optimized_tags> s(fcl);

  test_log_streambuf();
  test_cout();
  test_logger(s);
  test_all_accepted(s);
  test_all_rejected(s);
  test_all_rejected_compile_time(s);
  test_empty_loop(s);
//NOT USED  test_dead_code_elimination(s);
  return 0;
}
